"""
    symbolite.core.util
    ~~~~~~~~~~~~~~~~~~~

    Symbolite core util functions.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from types import ModuleType
from typing import Any, Callable, Hashable, Iterator, Mapping, TypeVar

from ..abstract import scalar
from . import evaluate, inspect, substitute

TH = TypeVar("TH", bound=Hashable)


def solve_dependencies(dependencies: Mapping[TH, set[TH]]) -> Iterator[set[TH]]:
    """Solve a dependency graph.

    Parameters
    ----------
    dependencies :
        dependency dictionary. For each key, the value is an iterable indicating its
        dependencies.

    Yields
    ------
    set
        iterator of sets, each containing keys of independents tasks dependent only of
        the previous tasks in the list.

    Raises
    ------
    ValueError
        if a cyclic dependency is found.
    """
    while dependencies:
        # values not in keys (items without dep)
        t = {i for v in dependencies.values() for i in v} - dependencies.keys()
        # and keys without value (items without dep)
        t.update(k for k, v in dependencies.items() if not v)
        # can be done right away
        if not t:
            raise ValueError(
                "Cyclic dependencies exist among these items: {}".format(
                    ", ".join(repr(x) for x in dependencies.items())
                )
            )
        # and cleaned up
        dependencies = {k: v - t for k, v in dependencies.items() if v}
        yield t


def compute_dependencies(
    content: Mapping[TH, Any],
    is_dependency: Callable[[Any], bool],
):
    dependencies = {}
    for k, v in content.items():
        contents = inspect(v)
        if contents == {k: 1}:
            dependencies[k] = set()
        else:
            dependencies[k] = set(filter(is_dependency, contents.keys()))
    return dependencies


def substitute_content(
    content: Mapping[TH, Any],
    *,
    is_dependency: Callable[[Any], bool],
) -> dict[TH, scalar.NumberT]:
    dependencies = compute_dependencies(content, is_dependency)
    layers = solve_dependencies(dependencies)

    out: dict[TH, Any] = {}
    for layer in layers:
        for item in layer:
            out[item] = substitute(content[item], out)

    return out


def eval_content(
    content: Mapping[TH, Any],
    *,
    libsl: ModuleType,
    is_dependency: Callable[[Any], bool],
) -> dict[TH, scalar.NumberT]:
    """Evaluate a group of

    Parameters
    ----------
    content
        a mapping of assigments.
    libsl
        symbolite implementation module.
    is_root
        callable that takes a python object/value and returns True
        if it should be considered as having no dependencies.
    is_dependency
        callable that takes a python object/value and returns True
        if it should be considered as the dependency of another.
    """
    dependencies = compute_dependencies(content, is_dependency)
    layers = solve_dependencies(dependencies)

    out: dict[TH, scalar.NumberT] = {}
    for layer in layers:
        for item in layer:
            out[item] = evaluate(substitute(content[item], out), libsl)

    return out



import math as pymath
import types
from operator import attrgetter
from typing import Any

import pytest

from symbolite.core import Unsupported
from symbolite.impl import get_all_implementations

all_impl = get_all_implementations()

# Available functions and their test values
_functions = {
    "abs": (3,),
    "acos": (0.5,),
    "acosh": (3,),
    "asin": (0.5,),
    "asinh": (3,),
    "atan": (3,),
    "atan2": (1, 2),
    "atanh": (0.5,),
    "ceil": (3,),
    "comb": (3, 2),
    "copysign": (1, 2),
    "cos": (3,),
    "cosh": (3,),
    "degrees": (3,),
    "erf": (3,),
    "erfc": (3,),
    "exp": (3,),
    "expm1": (3,),
    "fabs": (3,),
    "factorial": (3,),
    "floor": (3,),
    "fmod": (1, 2),
    "frexp": (3.0,),
    "gamma": (3,),
    # "gcd": (1, 2),
    # "hypot": (1, 2),
    # "isclose": (1, 2),
    "isfinite": (3,),
    "isinf": (3,),
    "isnan": (3,),
    "isqrt": (3,),
    # "lcm": (3, 2),
    "ldexp": (1, 2),
    "lgamma": (3,),
    "log": (3,),
    "log10": (3,),
    "log1p": (3,),
    "log2": (3,),
    "modf": (3.0,),
    "nextafter": (3, 4),
    # "perm": (3,),
    "pow": (1, 2),
    "radians": (3,),
    "remainder": (1, 2),
    "sin": (3,),
    "sinh": (3,),
    "sqrt": (3,),
    "tan": (3,),
    "tanh": (3,),
    "trunc": (3,),
    "ulp": (3,),
}

_values = {
    "e": None,
    "inf": None,
    "pi": None,
    "nan": None,
    "tau": None,
}

_all = {**_functions, **_values}


def almost_equal(x: Any, y: Any, threshold: float = 0.0001):
    if isinstance(x, bool):
        assert x == y
    elif isinstance(x, tuple):
        for el0, el1 in zip(x, y):
            almost_equal(el0, el1, threshold)
    else:
        if pymath.isnan(x):
            assert pymath.isnan(y)
        elif pymath.isinf(x):
            assert pymath.isinf(y)
        else:
            assert abs(x - y) < threshold, (x, "!=", y)


@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_is_defined(libsl: types.ModuleType):
    """Test that that all included libraries define members defined in symbolite.lib"""
    for k in tuple(_functions.keys()) + tuple(_values.keys()):
        name = f"scalar.{k}"
        try:
            attrgetter(name)(libsl)
        except AttributeError:
            assert False, f"{name} not defined."


@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
@pytest.mark.parametrize("func_name_and_values", _all.items(), ids=_all.keys())
def test_compare(libsl: types.ModuleType, func_name_and_values: tuple[Any, Any]):
    """Compare implementation for different mappers
    for a (very small) subset of values.
    """
    func_name, test_values = func_name_and_values

    try:
        original_func = getattr(pymath, func_name)
    except AttributeError:
        # This is a function not on math, but rather on builtins
        try:
            original_func = __builtins__[func_name]
        except KeyError:
            # This is a function not on math and not on builtins
            # It is a function added in later Python versions.
            return

    mapped_func = attrgetter("scalar." + func_name)(libsl)

    if mapped_func is Unsupported:
        return

    if test_values is None:
        almost_equal(original_func, mapped_func)
    elif isinstance(test_values, tuple):
        almost_equal(original_func(*test_values), mapped_func(*test_values))
    elif isinstance(test_values, list):
        almost_equal(original_func(test_values), mapped_func(test_values))
    else:
        assert False, "Unknown test values"



"""
    symbolite.impl.libstd.vector
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.vector
    into values and functions defined in Python's math module.

    :copyright: 2023 by Symbolite-array Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import math
import operator

from ...core import Unsupported

op_getitem = operator.getitem

sum = math.fsum
prod = math.prod

Vector = Unsupported



import pytest

from symbolite import Scalar, scalar
from symbolite.core import as_string

x, y, z = map(scalar.Scalar, "x y z".split())


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y, "x + y"),
        (x - y, "x - y"),
        (x * y, "x * y"),
        (x / y, "x / y"),
        (x**y, "x ** y"),
        (x // y, "x // y"),
        (((x**y) % z), "x ** y % z"),
    ],
)
def test_known_symbols(expr: Scalar, result: Scalar):
    assert as_string(expr) == result
    assert str(expr) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + scalar.cos(y), "x + scalar.cos(y)"),
        (x + scalar.pi, "x + scalar.pi"),
    ],
)
def test_lib_symbols(expr: Scalar, result: Scalar):
    assert as_string(expr) == result



"""
    symbolite.abstract.vector
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Objects and functions for vector operations.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from __future__ import annotations

import dataclasses
from typing import Any, Iterable, Mapping, Sequence, overload

from .scalar import NumberT, Scalar
from .symbol import BaseFunction, Symbol

VectorT = Iterable[NumberT]


@dataclasses.dataclass(frozen=True)
class Vector(Symbol):
    """A user defined symbol."""

    def __getitem__(self, key: int | Scalar) -> Scalar:
        return super().__getitem__(key)

    def __getattr__(self, key: Any):
        raise AttributeError(key)


@dataclasses.dataclass(frozen=True)
class CumulativeFunction(BaseFunction):
    namespace: str = "vector"
    arity: int = 1

    def __call__(self, arg1: Vector | VectorT) -> Scalar:
        return super()._call(arg1)  # type: ignore


sum = CumulativeFunction("sum", namespace="vector")
prod = CumulativeFunction("prod", namespace="vector")


@overload
def vectorize(
    expr: NumberT,
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> NumberT:
    ...


@overload
def vectorize(
    expr: Symbol,
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> Symbol:
    ...


@overload
def vectorize(
    expr: Iterable[NumberT | Symbol],
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[NumberT | Symbol, ...]:
    ...


def vectorize(
    expr: NumberT | Symbol | Iterable[NumberT | Symbol],
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> NumberT | Symbol | tuple[NumberT | Symbol, ...]:
    """Vectorize expression by replacing test_scalar symbols
    by an array at a given indices.

    Parameters
    ----------
    expr
    symbol_names
        if a tuple, provides the names of the symbols
        which will be mapped to the indices given by their position.
        if a dict, maps symbol names to indices.
    varname
        name of the array variable
    """
    if isinstance(expr, NumberT):
        return expr

    if not isinstance(expr, Symbol):
        return tuple(vectorize(symbol, symbol_names, varname) for symbol in expr)

    if isinstance(symbol_names, dict):
        it = zip(symbol_names.values(), symbol_names.keys())
    else:
        it = enumerate(symbol_names)

    arr = Vector(varname)

    reps = {scalar_type(name): arr[ndx] for ndx, name in it}
    return expr.subs(reps)


@overload
def auto_vectorize(
    expr: NumberT,
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], Symbol]:
    ...


@overload
def auto_vectorize(
    expr: Symbol,
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], Symbol]:
    ...


@overload
def auto_vectorize(
    expr: Iterable[Symbol],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], tuple[Symbol, ...]]:
    ...


def auto_vectorize(
    expr: NumberT | Symbol | Iterable[Symbol],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], NumberT | Symbol | tuple[NumberT | Symbol, ...]]:
    """Vectorize expression by replacing all test_scalar symbols
    by an array at a given indices. Symbols are ordered into
    the array alphabetically.

    Parameters
    ----------
    expr
    varname
        name of the array variable

    Returns
    -------
    tuple[str, ...]
        symbol names as ordered in the array.
    SymbolicExpression
        vectorized expression.
    """
    if isinstance(expr, NumberT):
        return tuple(), expr

    if not isinstance(expr, Symbol):
        expr = tuple(expr)
        out = set[str]()
        for symbol in expr:
            out.update(symbol.symbol_names(""))
        symbol_names = tuple(sorted(out))
        return symbol_names, vectorize(expr, symbol_names, varname, scalar_type)
    else:
        symbol_names = tuple(sorted(expr.symbol_names("")))
        return symbol_names, vectorize(expr, symbol_names, varname, scalar_type)



"""
    symbolite.impl.libjax.symbol
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.symbol
    into values and functions defined in JAX.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from ..libstd.symbol import *  # noqa: F403



import types
from typing import Any

import pytest

from symbolite import Symbol, scalar, vector
from symbolite.core import Unsupported
from symbolite.impl import get_all_implementations

all_impl = get_all_implementations()

x, y = map(scalar.Scalar, ("x", "y"))
vec = vector.Vector("vec")
v = vector.Vector("v")

xsy = Symbol("xsy")

requires_numpy = pytest.mark.skipif("libnumpy" not in all_impl, reason="Requires NumPy")
requires_sympy = pytest.mark.skipif("libsympy" not in all_impl, reason="Requires SymPy")


@pytest.mark.mypy_testing
# noqa: F821
def test_typing():
    reveal_type(v + v)  # R: symbolite.abstract.vector.Vector # noqa: F821
    reveal_type(2 + v)  # R: symbolite.abstract.vector.Vector # noqa: F821
    reveal_type(v + 2)  # R: symbolite.abstract.vector.Vector # noqa: F821
    # reveal_type(v + xsy) # R: symbolite.abstract.symbol.Symbol # noqa: F821
    reveal_type(xsy + v)  # R: symbolite.abstract.symbol.Symbol # noqa: F821
    reveal_type(vec[0])  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    reveal_type(vec[x])  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    reveal_type(vector.sum(vec))  # R: symbolite.abstract.scalar.Scalar # noqa: F821


def test_vector():
    assert str(vec) == "vec"
    assert str(vec[1]) == "vec[1]"


def test_methods():
    assert vec.subs_by_name(vec=(1, 2, 3)) == (1, 2, 3)
    assert vec[1].subs_by_name(vec=(1, 2, 3)).eval() == 2
    assert vec.symbol_names() == {
        "vec",
    }
    assert vec[1].symbol_names() == {
        "vec",
    }
    assert (vec[1] + vec[0]).symbol_names() == {
        "vec",
    }


@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_impl(libsl: types.ModuleType):
    v = vector.Vector("v")
    try:
        expr = vector.sum(v)
        assert expr.subs_by_name(v=(1, 2, 3, 4)).eval(libsl=libsl) == 10
    except Unsupported:
        pass

    expr = vector.prod(v)
    assert expr.subs_by_name(v=(1, 2, 3, 4)).eval(libsl=libsl) == 24


@requires_numpy
def test_impl_numpy():
    try:
        import numpy as np

        from symbolite.impl import libnumpy as libsl
    except ImportError:
        return

    v = np.asarray((1, 2, 3))

    expr1 = vector.Vector("vec") + 1
    assert np.allclose(expr1.subs_by_name(vec=v).eval(), v + 1)

    expr2 = scalar.cos(vector.sum(vector.Vector("vec")))

    assert np.allclose(expr2.subs_by_name(vec=v).eval(libsl=libsl), np.cos(np.sum(v)))


@requires_sympy
def test_impl_sympy():
    try:
        import sympy as sy

        from symbolite.impl import libsympy as libsl
    except ImportError:
        return

    vec = vector.Vector("vec")
    syarr = sy.IndexedBase("vec")
    assert vec.eval(libsl=libsl) == syarr
    assert vec[1].eval(libsl=libsl) == syarr[1]


@pytest.mark.parametrize(
    "expr,params,result",
    [
        (x + 2 * y, ("x", "y"), vec[0] + 2 * vec[1]),
        (x + 2 * y, ("y", "x"), vec[1] + 2 * vec[0]),
        (x + 2 * scalar.cos(y), ("y", "x"), vec[1] + 2 * scalar.cos(vec[0])),
        (x + 2 * y, dict(x=3, y=5), vec[3] + 2 * vec[5]),
        (x + 2 * y, dict(x=5, y=3), vec[5] + 2 * vec[3]),
    ],
)
def test_vectorize(expr: vector.Vector, params: Any, result: Symbol):
    assert vector.vectorize(expr, params) == result


def test_vectorize_non_default_varname():
    assert vector.vectorize(x + 2 * y, ("x", "y"), varname="v") == v[0] + 2 * v[1]


def test_vectorize_many():
    eqs = [
        x + 2 * y,
        y + 3 * x,
    ]
    result = (
        vec[2] + 2 * vec[0],
        vec[0] + 3 * vec[2],
    )
    assert vector.vectorize(eqs, ("y", "z", "x")) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + 2 * y, (("x", "y"), vec[0] + 2 * vec[1])),
        (y + 2 * x, (("x", "y"), vec[1] + 2 * vec[0])),
        (x + 2 * scalar.cos(y), (("x", "y"), vec[0] + 2 * scalar.cos(vec[1]))),
    ],
)
def test_autovectorize(expr: Symbol, result: Symbol):
    assert vector.auto_vectorize(expr) == result


def test_autovectorize_non_default_varname():
    assert vector.auto_vectorize(x + 2 * y, "v") == (("x", "y"), v[0] + 2 * v[1])


def test_autovectorize_many():
    eqs = [
        x + 2 * y,
        y + 3 * x,
    ]
    result = (
        vec[0] + 2 * vec[1],
        vec[1] + 3 * vec[0],
    )
    assert vector.auto_vectorize(eqs) == (("x", "y"), result)



import pickle

from pytest import mark

from .. import Scalar, Symbol, Vector


@mark.parametrize("cls", [Scalar, Symbol, Vector])
def test_pickle(cls):
    obj = cls()
    dump = pickle.dumps(obj)
    load = pickle.loads(dump)
    assert load == obj



"""
    symbolite.impl.libsympy.vector
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.vector
    into values and functions defined in SymPy.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import operator

import sympy as sy

op_getitem = operator.getitem

sum = sum
prod = sy.prod

Vector = sy.IndexedBase



"""
    symbolite.impl.libnumpy.scalar
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.scalar
    into values and functions defined in NumPy.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import numpy as np

from symbolite.core import Unsupported

abs = np.abs
acos = np.arccos
acosh = np.arccosh
asin = np.arcsin
asinh = np.arcsinh
atan = np.arctan
atan2 = np.arctan2
atanh = np.arctanh
ceil = np.ceil
comb = Unsupported
copysign = np.copysign
cos = np.cos
cosh = np.cosh
degrees = np.degrees
erf = Unsupported
erfc = Unsupported
exp = np.exp
expm1 = np.expm1
fabs = np.fabs
factorial = Unsupported
floor = np.floor
fmod = np.fmod
frexp = np.frexp
gamma = Unsupported
hypot = np.hypot
isfinite = np.isfinite
isinf = np.isinf
isnan = np.isnan
isqrt = Unsupported
ldexp = np.ldexp
lgamma = Unsupported
log = np.log
log10 = np.log10
log1p = np.log1p
log2 = np.log2
modf = np.modf
nextafter = np.nextafter
pow = np.power
radians = np.radians
remainder = np.remainder
sin = np.sin
sinh = np.sinh
sqrt = np.sqrt
tan = np.tan
tanh = np.tanh
trunc = np.trunc
ulp = Unsupported

e = np.e
inf = np.inf
pi = np.pi
nan = np.nan
tau = 2 * np.pi

Scalar = Unsupported

del np, Unsupported



import inspect
import types

import pytest

from symbolite import Symbol, scalar
from symbolite.core import as_function
from symbolite.impl import get_all_implementations

all_impl = get_all_implementations()

x, y, z = map(scalar.Scalar, "x y z".split())

xsy = Symbol("xsy")


@pytest.mark.mypy_testing
def test_typing():
    reveal_type(x + y)  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    reveal_type(2 + y)  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    reveal_type(x + 2)  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    # reveal_type(x + xsy) # R: symbolite.abstract.symbol.Symbol # noqa: F821
    # reveal_type(xsy + x) # R: symbolite.abstract.symbol.Symbol # noqa: F821
    reveal_type(scalar.cos(x))  # R: symbolite.abstract.scalar.Scalar # noqa: F821
    # reveal_type(scalar.cos(xsy)) # R: symbolite.abstract.scalar.Scalar # noqa: F821


@pytest.mark.parametrize(
    "expr",
    [
        x + y,
        x - y,
        x * y,
        x / y,
        x**y,
        x // y,
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_known_symbols(expr: Symbol, libsl: types.ModuleType):
    f = as_function(expr, "my_function", ("x", "y"), libsl=libsl)
    assert f.__name__ == "my_function"
    assert expr.subs_by_name(x=2, y=3).eval(libsl=libsl) == f(2, 3)
    assert tuple(inspect.signature(f).parameters.keys()) == ("x", "y")


@pytest.mark.parametrize(
    "expr,replaced",
    [
        (x + scalar.cos(y), 2 + scalar.cos(3)),
        (x + scalar.pi * y, 2 + scalar.pi * 3),
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_lib_symbols(expr: Symbol, replaced: Symbol, libsl: types.ModuleType):
    f = as_function(expr, "my_function", ("x", "y"), libsl=libsl)
    value = f(2, 3)
    assert f.__name__ == "my_function"
    assert expr.subs_by_name(x=2, y=3).eval(libsl=libsl) == value
    assert tuple(inspect.signature(f).parameters.keys()) == ("x", "y")


@pytest.mark.parametrize(
    "expr,namespace,result",
    [
        (
            x + scalar.pi * scalar.cos(y),
            None,
            {
                "x",
                "y",
                "scalar.cos",
                "scalar.pi",
                "symbol.mul",
                "symbol.add",
            },
        ),
        (x + scalar.pi * scalar.cos(y), "", {"x", "y"}),
        (
            x + scalar.pi * scalar.cos(y),
            "scalar",
            {"scalar.cos", "scalar.pi"},
        ),
    ],
)
def test_list_symbols(expr: Symbol, namespace: str | None, result: Symbol):
    assert expr.symbol_names(namespace) == result



import pytest

from symbolite import Scalar, scalar
from symbolite.core import as_string

x, y, z = map(scalar.Scalar, "x y z".split())


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y * z, "x + y * z"),
        ((x + y) * z, "(x + y) * z"),
        (x * y + z, "x * y + z"),
        (x * (y + z), "x * (y + z)"),
        (-(x**y), "-x ** y"),
        ((-x) ** y, "(-x) ** y"),
    ],
)
def test_different_precedence(expr: Scalar, result: Scalar):
    assert as_string(expr) == result
    assert str(expr) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y + z, "x + y + z"),
        ((x + y) + z, "x + y + z"),
        (x + (y + z), "x + (y + z)"),  # Python is not associative
    ],
)
def test_same_precedence(expr: Scalar, result: Scalar):
    assert as_string(expr) == result
    assert str(expr) == result



"""
    symbolite.abstract.symbol
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Objects and functions for symbol operations.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import dataclasses
import functools
import types
from operator import attrgetter
from typing import Any, Callable, Generator, Mapping

from typing_extensions import Self

from ..core import (
    Unsupported,
    evaluate,
    substitute,
    substitute_by_name,
)


def filter_namespace(
    namespace: str | None = "", include_anonymous: bool = False
) -> Callable[[Named], bool]:
    def _inner(s: Named) -> bool:
        if namespace is None:
            return True
        return s.namespace == namespace

    return _inner


@dataclasses.dataclass(frozen=True)
class Named:
    name: str | None = None
    namespace: str = ""

    def __str__(self) -> str:
        if self.name is None:
            return "<anonymous>"

        if self.namespace:
            return self.namespace + "." + self.name

        return self.name

    @property
    def is_anonymous(self) -> bool:
        return self.name is None

    def format(self, *args: Any, **kwargs: Any) -> str:
        ...


@dataclasses.dataclass(frozen=True)
class Symbol(Named):
    """Base class for objects that might operate with others using
    python operators that map to magic methods

    The following magic methods are not mapped to symbolite Functions
      - __hash__, __eq__, __ne__ collides with reasonble use of comparisons
        within user code (including uses as dict keys).
        We defined `.eq` y `.ne` methods for the two lasts.
      - __contains__ is coerced to boolean.
      - __bool__ yields a TypeError if not boolean.
      - __str__, __bytes__, __repr__ yields a TypeError if the return value
        is not of the corresponding type.
        and they might also affect usability in the console.
      - __format__
      - __int__, __float__, __complex__ yields a TypeError if the return value
        is not of the corresponding type.
      - __round__, __abs__, __divmod__ they are too "numeric related"
      - __trunc__, __ceil__, __floor__ they are too "numeric related"
        and called by functions in math.
      - __len__ yields a TypeError if not int.
      - __index__ yields a TypeError if not int.

    Also, magic methods that are statements (not expressions) are also not
    mapped: e.g. __setitem__ or __delitem__

    """

    expression: Expression | None = None

    # Comparison methods (not operator)
    def eq(self, other: Any) -> Self:
        return eq(self, other)

    def ne(self, other: Any) -> Self:
        return ne(self, other)

    # Comparison magic methods
    def __lt__(self, other: Any) -> Self:
        """Implements less than comparison using the < operator."""
        return lt(self, other)

    def __le__(self, other: Any) -> Self:
        """Implements less than or equal comparison using the <= operator."""
        return le(self, other)

    def __gt__(self, other: Any) -> Self:
        """Implements greater than comparison using the > operator."""
        return gt(self, other)

    def __ge__(self, other: Any) -> Self:
        """Implements greater than or equal comparison using the >= operator."""
        return ge(self, other)

    # Emulating container types
    def __getitem__(self, key: Any) -> Self:
        """Defines behavior for when an item is accessed,
        using the notation self[key]."""
        return getitem(self, key)

    # Emulating attribute
    def __getattr__(self, key: str) -> Self:
        """Defines behavior for when an item is accessed,
        using the notation self.key"""
        if key.startswith("__"):
            raise AttributeError(key)
        return symgetattr(self, key)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Self:
        """Implements addition."""
        return add(self, other)

    def __sub__(self, other: Any) -> Self:
        """Implements subtraction."""
        return sub(self, other)

    def __mul__(self, other: Any) -> Self:
        """Implements multiplication."""
        return mul(self, other)

    def __matmul__(self, other: Any) -> Self:
        """Implements multiplication."""
        return matmul(self, other)

    def __truediv__(self, other: Any) -> Self:
        """Implements true division."""
        return truediv(self, other)

    def __floordiv__(self, other: Any) -> Self:
        """Implements integer division using the // operator."""
        return floordiv(self, other)

    def __mod__(self, other: Any) -> Self:
        """Implements modulo using the % operator."""
        return mod(self, other)

    def __pow__(self, other: Any, modulo: Any = None) -> Self:
        """Implements behavior for exponents using the ** operator."""
        if modulo is None:
            return pow(self, other)
        else:
            return pow3(self, other, modulo)

    def __lshift__(self, other: Any) -> Self:
        """Implements left bitwise shift using the << operator."""
        return lshift(self, other)

    def __rshift__(self, other: Any) -> Self:
        """Implements right bitwise shift using the >> operator."""
        return rshift(self, other)

    def __and__(self, other: Any) -> Self:
        """Implements bitwise and using the & operator."""
        return and_(self, other)

    def __or__(self, other: Any) -> Self:
        """Implements bitwise or using the | operator."""
        return or_(self, other)

    def __xor__(self, other: Any) -> Self:
        """Implements bitwise xor using the ^ operator."""
        return xor(self, other)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Self:
        """Implements reflected addition."""
        return add(other, self)

    def __rsub__(self, other: Any) -> Self:
        """Implements reflected subtraction."""
        return sub(other, self)

    def __rmul__(self, other: Any) -> Self:
        """Implements reflected multiplication."""
        return mul(other, self)

    def __rmatmul__(self, other: Any) -> Self:
        """Implements reflected multiplication."""
        return matmul(other, self)

    def __rtruediv__(self, other: Any) -> Self:
        """Implements reflected true division."""
        return truediv(other, self)

    def __rfloordiv__(self, other: Any) -> Self:
        """Implements reflected integer division using the // operator."""
        return floordiv(other, self)

    def __rmod__(self, other: Any) -> Self:
        """Implements reflected modulo using the % operator."""
        return mod(other, self)

    def __rpow__(self, other: Any) -> Self:
        """Implements behavior for reflected exponents using the ** operator."""
        return pow(other, self)

    def __rlshift__(self, other: Any) -> Self:
        """Implements reflected left bitwise shift using the << operator."""
        return lshift(other, self)

    def __rrshift__(self, other: Any) -> Self:
        """Implements reflected right bitwise shift using the >> operator."""
        return rshift(other, self)

    def __rand__(self, other: Any) -> Self:
        """Implements reflected bitwise and using the & operator."""
        return and_(other, self)

    def __ror__(self, other: Any) -> Self:
        """Implements reflected bitwise or using the | operator."""
        return or_(other, self)

    def __rxor__(self, other: Any) -> Self:
        """Implements reflected bitwise xor using the ^ operator."""
        return xor(other, self)

    # Unary operators and functions
    def __neg__(self) -> Self:
        """Implements behavior for negation (e.g. -some_object)"""
        return neg(self)

    def __pos__(self) -> Self:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return pos(self)

    def __invert__(self) -> Self:
        """Implements behavior for inversion using the ~ operator."""
        return invert(self)

    def __str__(self) -> str:
        if self.expression is None:
            return super().__str__()
        return str(self.expression)

    def yield_named(
        self, include_anonymous: bool = False
    ) -> Generator[Named, None, None]:
        if self.expression is None:
            if include_anonymous or not self.is_anonymous:
                yield self
        else:
            yield from self.expression.yield_named(include_anonymous)

    def subs(self, mapper: Mapping[Any, Any]) -> Self:
        """Replace symbols, functions, values, etc by others.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        mappers
            dictionary mapping source to destination objects.
        """
        if self.expression is None:
            return mapper.get(self, self)
        return substitute(self.expression, mapper)

    def subs_by_name(self, **mapper: Any) -> Self:
        """Replace Symbols by values or objects, matching by name.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        **mapper
            keyword arguments connecting names to values.
        """
        if self.expression is None:
            return mapper.get(str(self), self)
        return substitute_by_name(self.expression, **mapper)

    def eval(self, libsl: types.ModuleType | None = None) -> Any:
        """Evaluate expression.

        If no implementation library is provided:
        1. 'libsl' will be looked up going back though the stack
           until is found.
        2. If still not found, the implementation using the python
           math module will be used (and a warning will be issued).

        Parameters
        ----------
        libs
            implementations
        """
        if libsl is None:
            return evaluate(self)

        if self.expression is not None:
            return evaluate(self.expression, libsl)

        if self.namespace:
            name = str(self)

            value = attrgetter(name)(libsl)

            if value is Unsupported:
                raise Unsupported(f"{name} is not supported in module {libsl.__name__}")

            return value
        else:
            # User defined symbol, txry to map the class
            name = (
                f"{self.__class__.__module__.split('.')[-1]}.{self.__class__.__name__}"
            )
            f = attrgetter(name)(libsl)

            if f is Unsupported:
                raise Unsupported(f"{name} is not supported in module {libsl.__name__}")

            return f(self.name)

    def symbol_namespaces(self) -> set[str]:
        """Return a set of symbol libraries"""
        return set(map(lambda s: s.namespace, self.yield_named(False)))

    def symbol_names(self, namespace: str | None = "") -> set[str]:
        """Return a set of symbol names (with full namespace indication).

        Parameters
        ----------
        namespace: str or None
            If None, all symbols will be returned independently of the namespace.
            If a string, will compare Symbol.namespace to that.
            Defaults to "" which is the namespace for user defined symbols.
        """
        ff = filter_namespace(namespace)
        return set(map(str, filter(ff, self.yield_named(False))))


@dataclasses.dataclass(frozen=True, kw_only=True)
class BaseFunction(Named):
    """A callable primitive that will return a call."""

    fmt: str | None = None
    arity: int | None = None

    @property
    def call(self) -> type[Expression]:
        return Expression

    @property
    def output_type(self):
        return Symbol

    def _call(self, *args: Any, **kwargs: Any) -> Symbol:
        return self.output_type(expression=self._build_resolver(*args, **kwargs))

    def _build_resolver(self, *args: Any, **kwargs: Any) -> Expression:
        if self.arity is None:
            return self.call(self, args, tuple(kwargs.items()))
        if kwargs:
            raise ValueError(
                "If arity is given, keyword arguments should not be provided."
            )
        if len(args) != self.arity:
            raise ValueError(
                f"Invalid number of arguments ({len(args)}), expected {self.arity}."
            )
        return self.call(self, args)

    def format(self, *args: Any, **kwargs: Any) -> str:
        if self.fmt:
            return self.fmt.format(*args, **kwargs)

        plain_args = args + tuple(f"{k}={v}" for k, v in kwargs.items())
        return f"{str(self)}({', '.join((str(v) for v in plain_args))})"


@dataclasses.dataclass(frozen=True)
class Function(BaseFunction):
    def __call__(self, *args: Any, **kwargs: Any) -> Symbol:
        return self._call(*args, **kwargs)


def _add_parenthesis(
    self: UnaryFunction | BinaryFunction,
    arg: UnaryFunction | BinaryFunction,
    *,
    right: bool,
) -> str:
    match arg:
        case Symbol(
            expression=Expression(
                func=UnaryFunction(precedence=p) | BinaryFunction(precedence=p)
            )
        ):
            if p < self.precedence or (right and p <= self.precedence):
                return f"({arg})"
    return str(arg)


@dataclasses.dataclass(frozen=True, kw_only=True)
class UnaryFunction(BaseFunction):
    arity: int = 1
    precedence: int

    def format(self, *args: Any, **kwargs: Any) -> str:
        (x,) = args
        x = _add_parenthesis(self, x, right=False)
        return super().format(x)

    def __call__(self, arg1: Symbol) -> Symbol:
        return self._call(arg1)


@dataclasses.dataclass(frozen=True, kw_only=True)
class BinaryFunction(BaseFunction):
    arity: int = 2
    precedence: int

    def format(self, *args: Any, **kwargs: Any) -> str:
        x, y = args
        x = _add_parenthesis(self, x, right=False)
        y = _add_parenthesis(self, y, right=True)
        return super().format(x, y)

    def __call__(self, arg1: Symbol, arg2: Symbol) -> Symbol:
        return self._call(arg1, arg2)


@dataclasses.dataclass(frozen=True)
class Expression:
    """A Function that has been called with certain arguments."""

    func: Named
    args: tuple[Any, ...]
    kwargs_items: tuple[tuple[str, Any], ...] = ()

    def __post_init__(self) -> None:
        if isinstance(self.kwargs_items, dict):
            object.__setattr__(self, "kwargs_items", tuple(self.kwargs_items.items()))

    @functools.cached_property
    def kwargs(self) -> dict[str, Any]:
        return dict(self.kwargs_items)

    def __str__(self) -> str:
        return self.func.format(*self.args, *self.kwargs)

    def yield_named(
        self, include_anonymous: bool = False
    ) -> Generator[Named, None, None]:
        if include_anonymous or not self.func.is_anonymous:
            yield self.func

        for arg in self.args:
            if isinstance(arg, Symbol):
                yield from arg.yield_named(include_anonymous)

        for _, v in self.kwargs_items:
            if isinstance(v, Symbol):
                yield from v.yield_named(include_anonymous)

    def subs(self, mapper: Mapping[Any, Any]) -> Self:
        """Replace symbols, functions, values, etc by others.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        mappers
            dictionary mapping source to destination objects.
        """
        func = mapper.get(self.func, self.func)
        args = tuple(substitute(arg, mapper) for arg in self.args)
        kwargs = {k: substitute(arg, mapper) for k, arg in self.kwargs_items}

        try:
            return func(*args, **kwargs)
        except Exception as ex:
            try:
                ex.add_note(f"While evaluating {func}(*{args}, **{kwargs}): {ex}")
            except AttributeError:
                pass
            raise ex

    def subs_by_name(self, **mapper: Any) -> Self:
        """Replace symbols, functions, values, etc by others.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        mappers
            dictionary mapping source to destination objects.
        """
        func = mapper.get(str(self.func), self.func)
        args = tuple(substitute_by_name(arg, **mapper) for arg in self.args)
        kwargs = {k: substitute_by_name(arg, **mapper) for k, arg in self.kwargs_items}

        try:
            return func(*args, **kwargs)
        except Exception as ex:
            try:
                ex.add_note(f"While evaluating {func}(*{args}, **{kwargs}): {ex}")
            except AttributeError:
                pass
            raise ex

    def eval(self, libsl: types.ModuleType | None = None) -> Any:
        """Evaluate expression.

        If no implementation library is provided:
        1. 'libsl' will be looked up going back though the stack
           until is found.
        2. If still not found, the implementation using the python
           math module will be used (and a warning will be issued).

        Parameters
        ----------
        libs
            implementations
        """
        func = attrgetter(str(self.func))(libsl)
        args = tuple(evaluate(arg, libsl) for arg in self.args)
        kwargs = {k: evaluate(arg, libsl) for k, arg in self.kwargs_items}

        try:
            return func(*args, **kwargs)
        except Exception as ex:
            try:
                ex.add_note(f"While evaluating {func}(*{args}, **{kwargs}): {ex}")
            except AttributeError:
                pass
            raise ex

    def symbol_names(self, namespace: str | None = "") -> set[str]:
        """Return a set of symbol names (with full namespace indication).

        Parameters
        ----------
        namespace: str or None
            If None, all symbols will be returned independently of the namespace.
            If a string, will compare Symbol.namespace to that.
            Defaults to "" which is the namespace for user defined symbols.
        """
        ff = filter_namespace(namespace)
        return set(map(str, filter(ff, self.yield_named(False))))


# Comparison methods (not operator)
eq = BinaryFunction("eq", "symbol", precedence=-5, fmt="{} == {}")
ne = BinaryFunction("ne", "symbol", precedence=-5, fmt="{} != {}")

# Comparison
lt = BinaryFunction("lt", "symbol", precedence=-5, fmt="{} < {}")
le = BinaryFunction("le", "symbol", precedence=-5, fmt="{} <= {}")
gt = BinaryFunction("gt", "symbol", precedence=-5, fmt="{} > {}")
ge = BinaryFunction("ge", "symbol", precedence=-5, fmt="{} >= {}")

# Emulating container types
getitem = BinaryFunction("getitem", "symbol", precedence=5, fmt="{}[{}]")

# Emulating attribute
symgetattr = BinaryFunction("symgetattr", "symbol", precedence=5, fmt="{}.{}")

# Emulating numeric types
add = BinaryFunction("add", "symbol", precedence=0, fmt="{} + {}")
sub = BinaryFunction("sub", "symbol", precedence=0, fmt="{} - {}")
mul = BinaryFunction("mul", "symbol", precedence=1, fmt="{} * {}")
matmul = BinaryFunction("matmul", "symbol", precedence=1, fmt="{} @ {}")
truediv = BinaryFunction("truediv", "symbol", precedence=1, fmt="{} / {}")
floordiv = BinaryFunction("floordiv", "symbol", precedence=1, fmt="{} // {}")
mod = BinaryFunction("mod", "symbol", precedence=1, fmt="{} % {}")
pow = BinaryFunction("pow", "symbol", precedence=3, fmt="{} ** {}")
pow3 = Function("pow3", "symbol", fmt="pow({}, {}, {})", arity=3)
lshift = BinaryFunction("lshift", "symbol", precedence=-1, fmt="{} << {}")
rshift = BinaryFunction("rshift", "symbol", precedence=-1, fmt="{} >> {}")
and_ = BinaryFunction("and_", "symbol", precedence=-2, fmt="{} & {}")
xor = BinaryFunction("xor", "symbol", precedence=-3, fmt="{} ^ {}")
or_ = BinaryFunction("or_", "symbol", precedence=-4, fmt="{} | {}")

# Unary operators
neg = UnaryFunction("neg", "symbol", precedence=2, fmt="-{}")
pos = UnaryFunction("pos", "symbol", precedence=2, fmt="+{}")
invert = UnaryFunction("invert", "symbol", precedence=2, fmt="~{}")
